# Same as dagster.utils.merge_dicts
def merge_dicts(*args) -> dict:
    """
    Returns a dictionary with with all the keys in all of the input dictionaries.

    If multiple input dictionaries have different values for the same key, the returned dictionary
    contains the value from the dictionary that comes latest in the list.
    """
    result: dict = {}
    for arg in args:
        result.update(arg)
    return result



# pylint: disable=unused-argument

import os
import re
import subprocess
from collections import Counter
from datetime import timedelta
from typing import Dict, List, Optional

import typer
import yaml
from typer import Typer

from ... import gql, ui
from ...config_utils import dagster_cloud_options
from ...core.graphql_client import GqlShimClient
from .sync_method import (
    MutagenSyncMethod,
    SandboxConnectionInfo,
    SyncMethod,
    SyncState,
    SyncedDirectory,
)
from .utils import get_current_display_timestamp

CONNECTION_INFO_GIVE_UP_TIME = timedelta(seconds=120)

app = Typer(help="Interface with your dev sandbox.")


def get_all_connection_info(client: GqlShimClient) -> Dict[str, SandboxConnectionInfo]:
    """
    Returns a mapping of location name to sandbox connection info for all locations.
    """
    workspace_entries = gql.fetch_workspace_entries(client)
    return {
        workspace_entry["locationName"]: SandboxConnectionInfo(
            username=workspace_entry["connectionInfo"]["username"],
            hostname=workspace_entry["connectionInfo"]["hostname"],
            port=workspace_entry["connectionInfo"]["port"],
        )
        for workspace_entry in workspace_entries
        if workspace_entry.get("connectionInfo")
    }


def launch_ssh(arguments: List[str]) -> None:
    os.execvp("ssh", ["ssh", *arguments])


def get_default_sync_method() -> SyncMethod:
    return MutagenSyncMethod()


def dev_deployment_client_or_error(
    base_client_url: str, organization: str, api_token: str
) -> GqlShimClient:
    default_client = gql.graphql_client_from_url(base_client_url, api_token)
    try:
        dev_depl = gql.get_dev_deployment_name(default_client)
    except Exception:
        raise ui.error(
            "No dev sandbox found for your user - you need to set one up before you can use this command."
        )
    # Current deployment is dev depl
    if not dev_depl:
        return default_client

    # Else, construct new gql client
    return gql.graphql_client_from_url(gql.url_from_config(organization, dev_depl), api_token)


@app.command(name="ssh")
@dagster_cloud_options(requires_url=True)
def ssh(
    organization: str,
    deployment: str,
    api_token: str,
    url: str,
    location_name=typer.Argument(
        ..., help="The name of the code location whose user code server to connect to."
    ),
) -> None:
    """
    Connects to the user code server hosting a given code location.
    """
    client = dev_deployment_client_or_error(url, organization, api_token)

    all_connection_info = get_all_connection_info(client)

    if not location_name in all_connection_info:
        raise ui.error(f"Cannot find sandbox user code server for location {location_name}.")

    connection_info = all_connection_info[location_name]

    # TODO forward all other arguments to SSH
    launch_ssh(
        ["-p", str(connection_info.port), f"{connection_info.username}@{connection_info.hostname}"]
    )


def find_locations_yaml(search_directory: str) -> Optional[str]:
    """
    Given a search directory, finds a locations.yaml file located
    in the search directory or a parent. Returns None if none is found.
    """
    curr_dir = search_directory

    while not os.path.exists(os.path.join(curr_dir, "locations.yaml")):
        parent_dir = os.path.dirname(curr_dir)
        if parent_dir == curr_dir:
            return None
        curr_dir = parent_dir
    return curr_dir


def get_directory_identifier(location_name: str, index: int) -> str:
    return f"{re.sub('[^0-9a-zA-Z-]+', '-', location_name)}-{index}"


def get_directories_to_sync(
    directory: str, all_connection_info: Dict[str, SandboxConnectionInfo]
) -> List[SyncedDirectory]:
    """
    Finds a set of code sync mappings defined in a locations.yaml file,
    combines with the sandbox connection info to create a list of directories to sync.
    """
    locations_yaml_dir = find_locations_yaml(directory)
    if not locations_yaml_dir:
        raise ValueError("No locations.yaml file found")

    directories_to_sync: List[SyncedDirectory] = []
    with open(os.path.join(locations_yaml_dir, "locations.yaml"), encoding="utf8") as f:
        locations = yaml.safe_load(f.read())["locations"]

        defined_locations = set(locations.keys())
        available_locations = set(all_connection_info.keys())

        if len(available_locations.intersection(defined_locations)) == 0:
            defined_locations_display = ", ".join(locations.keys()) or "(None)"
            available_locations_display = ", ".join(all_connection_info.keys()) or "(None)"
            raise ui.error(
                "No accessable sync locations found\n\n"
                f"Defined in locations.yaml: {defined_locations_display}\nAvailable code locations: {available_locations_display}"
            )

        if len(defined_locations - available_locations) > 0:
            not_synced = ", ".join(defined_locations - available_locations)
            ui.warn(
                f"Some locations defined in locations.yaml are not accessible to sync: {not_synced}\n"
            )

        if len(available_locations - defined_locations) > 0:
            not_synced = ", ".join(available_locations - defined_locations)
            ui.warn(
                f"Some available locations are not defined in locations.yaml, and will not be synced: {not_synced}\n"
            )

        for location_name, location in locations.items():
            if location_name in all_connection_info:
                connection_info = all_connection_info[location_name]
                for i, mapping in enumerate(location.get("code_sync", [])):
                    directories_to_sync.append(
                        SyncedDirectory(
                            identifier=get_directory_identifier(location_name, i),
                            location_name=location_name,
                            from_directory=str(
                                os.path.join(locations_yaml_dir, mapping.get("from"))
                            ),
                            to_directory=mapping.get("to"),
                            connection_info=connection_info,
                        )
                    )
    return directories_to_sync


def validate_host_connection(connection_info: SandboxConnectionInfo):
    """
    Validates that we can SSH into the given host, and prompts the user to add to
    known_hosts if need be.
    """
    subprocess.check_output(
        [
            "ssh",
            "-p",
            f"{connection_info.port}",
            "-o",
            "ConnectTimeout=10",
            f"{connection_info.username}@{connection_info.hostname}",
            "exit",
        ]
    )


def print_sync_state(state_map: Dict[str, SyncState], remove_previous: bool = True):
    """
    Given a mapping from directory identifiers to sync states, prints a summary of
    all sync states.
    """
    if remove_previous:
        ui.erase_previous_line(2)

    c = Counter(state_map.values())
    n_unavailable = c.get(SyncState.UNAVAILABLE, 0)
    n_syncing = c.get(SyncState.SYNCING, 0)
    n_synced = c.get(SyncState.SYNCED, 0)
    total = len(state_map.keys()) - n_unavailable

    if n_syncing > 0:
        ui.print(
            ui.yellow(
                f"[{get_current_display_timestamp()}] ({n_synced}/{total}) Syncing directories..."
            )
        )
    else:
        ui.print(ui.blue(f"[{get_current_display_timestamp()}] ({n_synced}/{total}) Sync complete"))

    if n_unavailable > 0:
        ui.print(ui.red(f"           {n_unavailable} directories unable to sync"))
    elif remove_previous:
        ui.print()


@app.command(name="sync")
@dagster_cloud_options(requires_url=True)
def code_sync(
    organization: str,
    deployment: str,
    api_token: str,
    url: str,
    verbose: bool = typer.Option(False, "--verbose", "-v", is_flag=True),
    very_verbose: bool = typer.Option(False, "--very-verbose", "-vv", is_flag=True),
):
    """
    Starts to synchronize code in the local environment with code in your cloud sandbox.
    """
    if verbose and very_verbose:
        raise ui.error(
            f"Cannot specify both {ui.as_code('--verbose')} and {ui.as_code('--very-verbose')} options"
        )
    verbosity_level = 2 if very_verbose else 1 if verbose else 0

    client = dev_deployment_client_or_error(url, organization, api_token)

    sync_method = get_default_sync_method()
    sync_method.preflight(verbosity_level)

    # Populate list of locations to sync
    all_connection_info = get_all_connection_info(client)
    for connection_info in all_connection_info.values():
        try:
            validate_host_connection(connection_info)
        except Exception:
            raise ui.error("Unable to establish connection")

    directories_to_sync = get_directories_to_sync(os.path.abspath(os.getcwd()), all_connection_info)

    ui.print("Enabling code sync...")

    for directory in directories_to_sync:
        sync_method.create_directory_sync(directory)

    identifier_location_name_map = {
        directory.identifier: directory.location_name for directory in directories_to_sync
    }
    try:
        states_map = {dir.identifier: SyncState.SYNCING for dir in directories_to_sync}

        print()
        print_sync_state(states_map, remove_previous=True)

        for identifier, new_state in sync_method.sync_loop():
            last_state = states_map[identifier]
            states_map[identifier] = new_state

            if new_state == SyncState.SYNCED and last_state != new_state:
                gql.reload_repo_location(client, identifier_location_name_map[identifier])

            if (
                new_state == SyncState.SYNCING
                or new_state == SyncState.UNAVAILABLE
                or last_state != new_state
            ):
                print_sync_state(states_map, remove_previous=verbosity_level == 0)

    except KeyboardInterrupt:
        pass
    finally:
        ui.print("Terminating sync...")
        for directory in directories_to_sync:
            sync_method.cleanup_directory_sync(directory.identifier)



import logging
import time
from contextlib import AbstractContextManager, ExitStack, contextmanager
from typing import Any, Dict, Optional

import requests
import urllib3
from packaging.version import Version, parse
from requests.adapters import HTTPAdapter
from requests.exceptions import HTTPError
from urllib3 import Retry

from ._utils import merge_dicts
from .errors import DagsterCloudHTTPError, DagsterCloudMaintenanceException, GraphQLStorageError
from .headers.impl import get_dagster_cloud_api_headers

DEFAULT_RETRIES = 6
RETRY_BACKOFF_FACTOR = 0.5
DEFAULT_TIMEOUT = 60

logger = logging.getLogger("dagster_cloud")


class GqlShimClient(AbstractContextManager):
    """Adapter for gql.Client that wraps errors in human-readable format."""

    def __init__(
        self,
        url: str,
        headers: Optional[Dict[str, Any]] = None,
        verify: bool = True,
        retries: int = 1,
        timeout: int = DEFAULT_TIMEOUT,
        cookies: Optional[Dict[str, Any]] = None,
        session: Optional[requests.Session] = None,
    ):
        self._exit_stack = ExitStack()

        self.url = url
        self.headers = headers
        self.verify = verify
        self.retries = retries
        self.timeout = timeout
        self.cookies = cookies
        self._session = (
            session
            if session
            else self._exit_stack.enter_context(create_cloud_requests_session(self.retries))
        )

    @property
    def session(self) -> requests.Session:
        return self._session

    def execute(self, query: str, variable_values: Dict[str, Any] = None):

        start_time = time.time()

        while True:
            try:
                return self._execute_retry(query, variable_values)
            except DagsterCloudMaintenanceException as e:
                if time.time() - start_time > e.timeout:
                    raise

                logger.warning(
                    "Dagster Cloud is currently unavailable due to scheduled maintenance. Retrying in {retry_interval} seconds...".format(
                        retry_interval=e.retry_interval
                    )
                )
                time.sleep(e.retry_interval)

    def _execute_retry(self, query: str, variable_values: Dict[str, Any] = None):

        post_args = {
            "headers": merge_dicts(self.headers, {"Content-type": "application/json"}),
            "cookies": self.cookies,
            "timeout": self.timeout,
            "verify": self.verify,
            "json": merge_dicts(
                {
                    "query": query,
                },
                {"variables": variable_values} if variable_values else {},
            ),
        }

        try:
            response = self._session.post(self.url, **post_args)
            try:
                result = response.json()
                if not isinstance(result, dict):
                    result = {}
            except ValueError:
                result = {}

            if "errors" not in result and "data" not in result and "maintenance" not in result:
                response.raise_for_status()
                raise requests.HTTPError("Unexpected GraphQL response", response=response)

        except HTTPError as http_error:
            raise DagsterCloudHTTPError(http_error) from http_error
        except Exception as exc:
            raise GraphQLStorageError(exc.__str__()) from exc

        if "maintenance" in result:
            maintenance_info = result["maintenance"]
            raise DagsterCloudMaintenanceException(
                message=maintenance_info.get("message"),
                timeout=maintenance_info.get("timeout"),
                retry_interval=maintenance_info.get("retry_interval"),
            )

        if "errors" in result:
            raise GraphQLStorageError(f"Error in GraphQL response: {str(result['errors'])}")
        else:
            return result

    def __enter__(self):
        return self

    def __exit__(self, _exception_type, _exception_value, _traceback):
        self._exit_stack.close()


def get_agent_headers(config_value: Dict[str, Any]):
    return get_dagster_cloud_api_headers(
        config_value["agent_token"],
        deployment_name=config_value.get("deployment"),
        additional_headers=config_value.get("headers"),
    )


@contextmanager
def create_cloud_requests_session(retries: int):
    with requests.Session() as session:
        urllib_version = parse(urllib3.__version__)  # type: ignore[attr-defined]
        # method for whitelisting all methods (GET/POST/etc.) is moving from method_whitelist
        # to allowed_methods
        allowed_method_param = (
            {"allowed_methods": None}
            if urllib_version >= Version("1.26.0")
            else {"method_whitelist": None}
        )

        adapter = HTTPAdapter(
            max_retries=Retry(  # pylint: disable=unexpected-keyword-arg
                total=retries,
                backoff_factor=RETRY_BACKOFF_FACTOR,
                status_forcelist=[500, 502, 503, 504],
                **allowed_method_param,  # type: ignore
            )
        )
        session.mount("https://", adapter)
        session.mount("http://", adapter)
        yield session


def create_proxy_client(
    url: str, config_value: Dict[str, Any], session: Optional[requests.Session] = None
):
    return GqlShimClient(
        url=url,
        headers=merge_dicts({"Content-type": "application/json"}, get_agent_headers(config_value)),
        verify=config_value.get("verify", True),
        retries=config_value.get("retries", DEFAULT_RETRIES),
        timeout=config_value.get("timeout", DEFAULT_TIMEOUT),
        cookies=config_value.get("cookies", {}),
        session=session,
    )


def create_cloud_dagit_client(url: str, api_token: str, retries=3):
    return GqlShimClient(
        url=f"{url}/graphql",
        headers=get_dagster_cloud_api_headers(api_token),
        retries=retries,
    )



from requests import HTTPError


class GraphQLStorageError(Exception):
    """Raise this when there's an error in the GraphQL layer"""


DEFAULT_MAINTENANCE_TIMEOUT = 3600 * 4 * 4
DEFAULT_MAINTENANCE_RETRY_INTERVAL = 30


class DagsterCloudMaintenanceException(Exception):
    def __init__(self, message, timeout, retry_interval):
        self.timeout = timeout or DEFAULT_MAINTENANCE_TIMEOUT
        self.retry_interval = retry_interval or DEFAULT_MAINTENANCE_RETRY_INTERVAL
        super(DagsterCloudMaintenanceException, self).__init__(message)


class DagsterCloudHTTPError(Exception):
    """
    Clearer error message for exceptions hitting Dagster Cloud servers.
    """

    def __init__(self, http_error: HTTPError):
        self.response = http_error.response
        error_content = http_error.response.content.decode("utf-8", errors="ignore")
        super(DagsterCloudHTTPError, self).__init__(
            http_error.__str__() + ": " + str(error_content)
        )


def raise_http_error(response):
    try:
        response.raise_for_status()
    except HTTPError as e:
        raise DagsterCloudHTTPError(e) from e



from typing import Dict

from setuptools import find_packages, setup


def get_version() -> str:
    version: Dict[str, str] = {}
    with open("dagster_cloud_cli/version.py", encoding="utf8") as fp:
        exec(fp.read(), version)  # pylint: disable=W0122

    return version["__version__"]


setup(
    name="dagster_cloud_cli",
    version=get_version(),
    author_email="hello@elementl.com",
    packages=find_packages(exclude=["dagster_cloud.cli_tests*"]),
    include_package_data=True,
    install_requires=[
        "packaging>=20.9",
        "questionary",
        "requests",
        "typer[all]",
        "PyYAML>=5.1",
    ],
    extras_require={},
    author="Elementl",
    license="Apache-2.0",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "dagster-cloud = dagster_cloud_cli.entrypoint:app",
        ]
    },
)
